/**
 * The MIT License (MIT)
 *
 * Copyright (c) 2019 Erik Moqvist
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/**
 * This file was generated by pbtools version 0.1.0 Thu Jan 24 08:14:06 2019.
 */

#ifndef ADDRESS_BOOK_H
#define ADDRESS_BOOK_H

#include <stdint.h>
#include <stdbool.h>

#define EBADWIRETYPE                                   1
#define EOUTOFDATA                                     2

struct address_book_heap_t {
    char *buf_p;
    int size;
    int pos;
};

/**
 * Message Person in package address_book.
 */
enum address_book_person_phone_type_e {
    address_book_person_phone_type_mobile_e = 0,
    address_book_person_phone_type_home_e = 1,
    address_book_person_phone_type_work_e = 2
};

struct address_book_person_phone_number_t {
    char *number_p;
    enum address_book_person_phone_type_e type;
};

struct address_book_person_t {
    struct address_book_heap_t *heap_p;
    char *name_p;
    int32_t id;
    char *email_p;
    struct {
        int length;
        struct address_book_person_phone_number_t *items_p;
    } phones;
};

/**
 * Message AddressBook in package address_book.
 */
struct address_book_address_book_t {
    struct address_book_heap_t *heap_p;
    struct {
        int length;
        struct address_book_person_t *items_p;
    } people;
};

/**
 * Craete a new person in given workspace.
 *
 * @param[in] src_p Data to decode.
 * @param[in] size Size of src_p.
 *
 * @return Initialized address book, or NULL on failure.
 */
struct address_book_person_t *address_book_person_new(
    void *workspace_p,
    size_t size);

/**
 * Allocate given number of phone numbers.
 *
 * @param[in] src_p Data to decode.
 * @param[in] size Size of src_p.
 *
 * @return zero(0) or negative error code.
 */
int address_book_person_phones_alloc(
    struct address_book_person_t *person_p,
    int length);

/**
 * Encode message AddressBook defined in package address_book.
 *
 * @param[out] dst_p Buffer to encode into.
 * @param[in] size Size of dst_p.
 * @param[in] src_p Data to encode.
 *
 * @return Encoded data length or negative error code.
 */
int address_book_person_encode(
    struct address_book_person_t *person_p,
    uint8_t *encoded_p,
    size_t size);

/**
 * Decode message AddressBook defined in package address_book.
 *
 * @param[out] dst_p Decoded data.
 * @param[in] src_p Data to decode.
 * @param[in] size Size of src_p.
 *
 * @return Number of bytes decoded or negative error code.
 */
int address_book_person_decode(
    struct address_book_person_t *person_p,
    const uint8_t *encoded_p,
    size_t size);

/**
 * Craete a new address book in given workspace.
 *
 * @param[in] src_p Data to decode.
 * @param[in] size Size of src_p.
 *
 * @return Initialized address book, or NULL on failure.
 */
struct address_book_address_book_t *address_book_address_book_new(
    void *workspace_p,
    size_t size);

/**
 * Allocate given number of persons.
 *
 * @param[in] src_p Data to decode.
 * @param[in] size Size of src_p.
 *
 * @return zero(0) or negative error code.
 */
int address_book_address_book_people_alloc(
    struct address_book_address_book_t *address_book_p,
    int length);

/**
 * Encode message AddressBook defined in package address_book.
 *
 * @param[out] dst_p Buffer to encode into.
 * @param[in] size Size of dst_p.
 * @param[in] src_p Data to encode.
 *
 * @return Encoded data length or negative error code.
 */
int address_book_address_book_encode(
    struct address_book_address_book_t *address_book_p,
    uint8_t *encoded_p,
    size_t size);

/**
 * Decode message AddressBook defined in package address_book.
 *
 * @param[out] dst_p Decoded data.
 * @param[in] src_p Data to decode.
 * @param[in] size Size of src_p.
 *
 * @return Number of bytes decoded or negative error code.
 */
int address_book_address_book_decode(
    struct address_book_address_book_t *address_book_p,
    const uint8_t *encoded_p,
    size_t size);

#endif
